///=======================================================================
// This program is useful in dumping out the information from a frame
// file.
//
// Currently supported modes are hex and post processed.
//
// :TODO: Currently the Version 6 frames are supported in the post
// :TODO:   process mode.
//
//=======================================================================
#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <iomanip>
#include <iostream>
#include <list>
#include <memory>
#include <stdexcept>
#include <string>
#include <vector>

#include "general/AtExit.hh"
#include "general/fstream.hh"
#include "general/hash_map.hh"
#include "general/regex.hh"
#include "general/regexmatch.hh"
#include "general/reverse.hh"
#include "general/types.hh"

typedef INT_2U class_id_type;
typedef INT_2U version_type;

class FrHeader;
class FrObject;
class SymbolTable;

static INT_8U
ClassSize( const std::string& Class )
{
  INT_8U retval = 0;

  if ( Class == "CHAR" )
  {
    retval = sizeof( CHAR );
  }
  else if ( Class == "CHAR_U" )
  {
    retval = sizeof( CHAR_U );
  }
  else if ( Class == "INT_2S" )
  {
    retval = sizeof( INT_2S );
  }
  else if ( Class == "INT_2U" )
  {
    retval = sizeof( INT_2U );
  }
  else if ( Class == "INT_4S" )
  {
    retval = sizeof( INT_4S );
  }
  else if ( Class == "INT_4U" )
  {
    retval = sizeof( INT_4U );
  }
  else if ( Class == "INT_8S" )
  {
    retval = sizeof( INT_8S );
  }
  else if ( Class == "INT_8U" )
  {
    retval = sizeof( INT_8U );
  }
  else if ( Class == "REAL_4" )
  {
    retval = sizeof( REAL_4 );
  }
  else if ( Class == "REAL_8" )
  {
    retval = sizeof( REAL_8 );
  }
  return retval;
}

class Description
{
public:
  typedef std::vector< void* > struct_data_type;

  void AddSE( const std::string& Name,
	      const std::string& Class,
	      const std::string& Comment );

  inline std::string
  Class( INT_4U Index ) const
  {
    std::string	retval;

    if ( ( Index > 0 )
	 && ( Index < m_se.size( ) ) )
    {
      ses_type::const_iterator	cur = m_se.begin( );
      while( Index-- )
      {
	++cur;
      }
      retval = cur->s_class;
    }
    return retval;
  }

  inline void
  ClassName( const std::string& ClassName )
  {
    m_class_name = ClassName;
  }

  inline const std::string&
  ClassName( ) const
  {
    return m_class_name;
  }

  void Dump( const struct_data_type& Data ) const;

  bool Read( FrHeader& Header, struct_data_type& Data ) const;


private:
  friend class SymbolTable;

  struct info {
    std::string	s_name;
    std::string s_class;
    std::string s_comment;
  };

  typedef std::list< info >	ses_type;

  std::string	m_class_name;
  ses_type	m_se;

  void* read( FrHeader& Header, const std::string& Class );
};

class FrameStream
{
public:
  typedef INT_8U position_type;
  FrameStream( );

  std::istream& Advance( position_type Position );

  std::istream& AdvanceABS( position_type Position );

  inline position_type Position( ) const
  {
    return m_position;
  }

  std::istream& Read( void* Buffer, position_type Size );

  std::istream& operator()( std::istream& Stream );

private:
  std::istream*	m_stream;
  position_type	m_position;
};

class FrHeader
{
public:
  typedef INT_8U streampos_type;

  void operator( )( );

  void AddDescription( class_id_type Id, const Description& Description );

  void AppendDescription( const FrObject& Object, bool Pre );

  inline bool
  ByteSwapping( ) const
  {
    return m_byte_swapping;
  }

  void Dump( ) const;

  std::string String( const std::string& Class, void* Data ) const;

  const Description& GetDescription( class_id_type Id ) const;

  void* Read( const std::string& Class,
	      const SymbolTable& Symbols );

  inline INT_2U Version( ) const
  {
    return m_data.base.version;
  }

private:
  enum {
    T_UNKNOWN,
    T1
  };

  struct base_type {
    char igwd[ 5 ];
    unsigned char version;
  };

  struct array_type {
  public:
    static bool Match( const std::string& Class );
    std::string Print( ) const;
    static array_type* Read( const version_type Version,
			     const std::string& Class,
			     const SymbolTable& Symbols );

  private:
    array_type( );

    static Regex	pattern;
  };

  struct ptr_struct_type {
  public:
    static bool Match( const std::string& Class );
    std::string Print( ) const;
    static ptr_struct_type* Read( const version_type Version );

  private:
    ptr_struct_type( INT_2U Class, INT_4U Instnace, version_type Version );

    const INT_2U	m_class;
    const INT_4U	m_instance;
    const version_type	m_version;
  };

  struct string_type {
  public:
    std::string Print( ) const;
    static bool Match( const std::string& Class );
    static string_type* Read( const version_type Version );

  private:
    string_type( const INT_2U Length, const char* Data,
		 const version_type Version );

    const INT_2U	m_len;
    const char* const	m_data;
    const version_type	m_version;
  };

  union {
    base_type base;
    struct {
      base_type	m_base;
      unsigned char minorVersion;
      unsigned char sizeInt2;
      unsigned char sizeInt4;
      unsigned char sizeInt8;
      unsigned char sizeReal4;
      unsigned char sizeReal8;
    
      unsigned char x1234[ 2 ];
      unsigned char x12345678[ 4 ];
      unsigned char x123456789abcdef[ 8 ];
        
      unsigned char smallPi[ 4 ];
      unsigned char bigPi[ 8 ];
    
      unsigned char AZ[ 2 ];
    } t1;
  } m_data;

  typedef General::hash_map< class_id_type, Description >	structs_type;

  structs_type		m_structs;

  int m_type;
  bool			m_byte_swapping;

};

class FrObject
{
public:
  FrObject( );

  ~FrObject( );

  void AppendSE( Description& D ) const;

  class_id_type	StructId( ) const;

  class_id_type	ClassId( ) const;

  std::string ClassName( ) const;

  void Dump( ) const;

  void Read( );

private:
  class FrBaseCore
  {
  public:
    typedef INT_8U size_type;

    virtual ~FrBaseCore( );

    virtual void Dump( ) const = 0;

    virtual size_type Bytes( ) const = 0;

    virtual class_id_type Class( ) const = 0;

    virtual size_type Length( ) const = 0;

    void Read( FrHeader& Header );
  };

  template< typename LT, typename CT, typename IT >
  class FrBase
    : public FrBaseCore
  {
  public:
    FrBase( FrHeader& Header );

    virtual size_type Bytes( ) const;

    virtual void Dump( ) const;

    virtual class_id_type Class( ) const;

    virtual size_type Length( ) const;

  private:
    LT	m_length;
    CT	m_class;
    IT	m_instance;
  }; // class - FrBase

  std::auto_ptr< FrBaseCore >		m_core;
  Description::struct_data_type		m_data;
  FrameStream::position_type		m_start_pos;
};

enum {
  MODE_STRUCT,
  MODE_HEX
};

class Output
{
public:
  Output( std::ostream& Stream );

  template < class T > std::ostream& operator<<( const T& Data );

private:
  std::ostream& m_stream;
};

class SymbolTable
{
public:
  typedef const Description::struct_data_type& data_type;
  typedef const Description::ses_type& data_desc_type;
  SymbolTable( data_desc_type DataDesc, data_type Values )
    : m_data_desc( DataDesc ),
      m_data( Values )
  {
  }

  INT_4U size( ) const
  {
    return m_data_desc.size( );
  }

  INT_8U LookupInteger( const std::string& Key ) const
  {
    INT_8U retval = 0;
    Description::struct_data_type::const_iterator
      symbol = m_data.begin( );

    std::istringstream m( Key );
    m >> retval;

    if ( retval == 0 )
    {
      for ( Description::ses_type::const_iterator
	      cur = m_data_desc.begin( ),
	      last = m_data_desc.end( );
	    cur != last;
	    ++cur, ++symbol )
      {
	if ( cur->s_name == Key )
	{
	  if ( cur->s_class == "CHAR" )
	  {
	    retval = *(static_cast<CHAR*>(*symbol));
	  }
	  else if ( cur->s_class == "CHAR_U" )
	  {
	    retval = *(static_cast<CHAR_U*>(*symbol));
	  }
	  else if ( cur->s_class == "INT_2S" )
	  {
	    retval = *(static_cast<INT_2S*>(*symbol));
	  }
	  else if ( cur->s_class == "INT_2U" )
	  {
	    retval = *(static_cast<INT_2U*>(*symbol));
	  }
	  else if ( cur->s_class == "INT_4S" )
	  {
	    retval = *(static_cast<INT_4S*>(*symbol));
	  }
	  else if ( cur->s_class == "INT_4U" )
	  {
	    retval = *(static_cast<INT_4U*>(*symbol));
	  }
	  else if ( cur->s_class == "INT_8S" )
	  {
	    retval = *(static_cast<INT_8S*>(*symbol));
	  }
	  else if ( cur->s_class == "INT_8U" )
	  {
	    retval = *(static_cast<INT_8U*>(*symbol));
	  }
	  break;
	}
      }
    }
    return retval;
  }

private:
  data_desc_type	m_data_desc;
  data_type		m_data;
};

static const char* Program;
static bool Scientific = false;
static bool VerboseData = true;
static bool Verbose = true;
static int Mode = MODE_STRUCT;

static FrameStream	FStream;
static FrHeader		Header;
static Output		Log( std::cout );

static void usage( )
{
  std::cerr << "Usage: "
	    << Program
	    << " [-e] [-S] [-s] [-h] [-m hex|struct]"
	    << " frame_file [ frame_file ... ]" << std::endl
	    << "\t-e - Print real numbers in scientific notation"
	    << std::endl
	    << "\t-S - silent with respect to the data" << std::endl
	    << "\t-s - silent mode " << std::endl
	    << "\t-h - display this help message" << std::endl
	    << "\t-m <mode> - specify the mode" << std::endl
	    << std::endl;
}

inline void
depart( int Status )
{
  General::AtExit::Cleanup( );
  exit( Status );
}

int
main( int ArgC, char**ArgV )
try
{
  Program = ArgV[ 0 ];

  int opt;

  while( ( opt = getopt( ArgC, ArgV, "ehm:Ss" ) ) != -1 )
  {
    switch( opt )
    {
    case 'e': /* scientific notation */
      Scientific = true;
      break;
    case 'h':	/* help */
      depart( 1 );
      break;
    case 'S': /* silent with respect to the data */
      VerboseData = false;
      break;
    case 'm':
      if ( strcmp( optarg, "hex" ) == 0 )
      {
	Mode = MODE_HEX;
      }
      else if ( strcmp( optarg, "struct" ) == 0 )
      {
	Mode = MODE_STRUCT;
      }
      break;
    case 's':	/* silent */
      Verbose = false;
      break;
    default:
      break;
    }
  }

  ArgC -= optind;
  ArgV += optind;

  if ( ArgC < 1 )
  {
    usage( );
    depart( 1 );
  }
  for ( int x = 0; x < ArgC; x++ )
  {
    //-------------------------------------------------------------------
    // Open the stream
    //-------------------------------------------------------------------
    General::ifstream ifile( ArgV[ x ] );

    if ( ! ifile.good( ) )
    {
      continue;
    }

    //-------------------------------------------------------------------
    // Read the FrHeader structure and write out the information
    //-------------------------------------------------------------------
    FStream( ifile );
    Header( );

    Header.Dump( );

    //-------------------------------------------------------------------
    // Read and then write all of the next elements.
    //-------------------------------------------------------------------
    try {
      while ( 1 )
      {
	FrObject o;

	if ( ifile.eof( ) || ( ! ifile.good( ) ) )
	{
	  break;
	}
	o.Dump( );
      }
    }
    catch( const std::exception& e )
    {
      //-----------------------------------------------------------------
      // Report failure back to user
      //-----------------------------------------------------------------
      std::cerr << "FAILURE: " << e.what( ) << std::endl;
    }
    catch( ... )
    {
      //-----------------------------------------------------------------
      // An unexpected error occured while reading the file.
      // Abandon further reading of the file.
      //-----------------------------------------------------------------
      std::cerr << "FAILURE: unknown exception" << std::endl;
    }
    //-------------------------------------------------------------------
    // Close the stream
    //-------------------------------------------------------------------
    ifile.close( );
  }
  depart( 0 );
}
catch( std::exception& e )
{
  std::cerr << "ABORT: Caught exception: " << e.what( ) << std::endl;
  depart( 1 );
}


//=======================================================================
// Description class definitions
//=======================================================================

void Description::
AddSE( const std::string& Name,
       const std::string& Class,
       const std::string& Comment )
{
  info	i;

  i.s_name = Name;
  i.s_class = Class;
  i.s_comment = Comment;

  m_se.push_back( i );
}

void Description::
Dump( const Description::struct_data_type& Data ) const
{
  void* const * retval_pos = &(Data[0]);
  
  for ( ses_type::const_iterator
	  cur = m_se.begin( ),
	  last = m_se.end( );
	cur != last;
	++cur, ++retval_pos )
  {
    Log << "  " << cur->s_name << ": "
	<< Header.String( cur->s_class, *retval_pos )
	<< std::endl;
  }
}

bool Description::
Read( FrHeader& Header, struct_data_type& Data ) const
{
  Data.resize( m_se.size( ) );
  void** pos = &((Data)[0]);

  SymbolTable	symbols( m_se, Data );
  bool		complete( true );

  ses_type::const_iterator
    cur = m_se.begin( ),
    last = m_se.end( );
  while ( cur != last )
  {
   try
   {
     *pos = Header.Read( cur->s_class, symbols );
     ++cur;
     ++pos;
   }
   catch( ... )
   {
     complete = false;
     break;
   }
  }
  if ( ! complete )
  {
    while( cur != last )
    {
      *pos = (void*)NULL;
      ++cur;
      ++pos;
    }
  }

  return complete;
}

//=======================================================================
// FrameStream class definition
//=======================================================================

FrameStream::
FrameStream( )
{
}

std::istream& FrameStream::
Advance( position_type Position )
{
  m_stream->seekg( Position, std::ios_base::cur );
  m_position += Position;
  return *m_stream;
}

std::istream& FrameStream::
AdvanceABS( position_type Position )
{
  m_stream->seekg( Position, std::ios_base::beg );
  m_position = Position;
  return *m_stream;
}

std::istream& FrameStream::
Read( void* Buffer, position_type Size )
{
  m_stream->read( (char*)Buffer, Size );

  m_position += m_stream->gcount( );

  return *m_stream;
}

std::istream& FrameStream::
operator( )( std::istream& Stream )
{
  m_stream = &Stream;
  m_position = position_type( 0 );
  return *m_stream;
}

//=======================================================================
// FrHeader class definitions
//=======================================================================

void FrHeader::
operator( )( )
{
  FStream.Read( &(m_data.base ), sizeof( m_data.base ) );
  //---------------------------------------------------------------------
  // 
  //---------------------------------------------------------------------
  if ( m_data.base.version >= 3 )
  {
    Description sh;
    Description se;

    sh.ClassName( "FrSH" );
    sh.AddSE( "name", "STRING", "" );
    sh.AddSE( "class", "INT_2U", "" );
    sh.AddSE( "comment", "STRING", "" );

    se.ClassName( "FrSE" );
    se.AddSE( "name", "STRING", "" );
    se.AddSE( "class", "STRING", "" );
    se.AddSE( "comment", "STRING", "" );

    AddDescription( 1, sh );
    AddDescription( 2, se );
    m_type = T1;
  }
  else
  {
    m_type = T_UNKNOWN;
  }

  switch( m_type )
  {
  case T1:
    //-------------------------------------------------------------------
    // Read the remainder of the FrHeader structure
    //-------------------------------------------------------------------
    FStream.Read( (void*)( (char*)(&(m_data.t1))
			   + sizeof( m_data.base) ),
		  sizeof( m_data.t1) - sizeof( m_data.base ) );
    //-------------------------------------------------------------------
    // Determinate byte ordering
    //---------------------------------------------------------------------
    INT_8U	sx123456789abcdef;

    ::memcpy( &sx123456789abcdef, m_data.t1.x123456789abcdef,
	      sizeof( sx123456789abcdef ) );
    m_byte_swapping = ( sx123456789abcdef != 0x123456789abcdefULL );

    break;
  }
}

void FrHeader::
AddDescription( class_id_type Id, const Description& Description )
{
  m_structs[ Id ] = Description;
}

void FrHeader::
AppendDescription( const FrObject& Object, bool Pre )
{
  static class_id_type			id = 0;
  static std::auto_ptr< Description >	d( (Description*)NULL );

  if ( Pre )
  {
    if ( ( Object.StructId( ) > 2 )
	 && ( d.get( ) ) )
    {
      AddDescription( id, *d );
      id = 0;
      d.reset( (Description*)NULL );
    }
  }
  else
  {
    if ( Object.StructId( ) == 1 )
    {
      if ( d.get( ) != (Description*)NULL )
      {
	AddDescription( id, *d );
      }
      d.reset( new Description );
      id = Object.ClassId( );
      d->ClassName( Object.ClassName( ) );
    }
    else if ( Object.StructId( ) == 2 )
    {
      Object.AppendSE( *d );
    }
  }
}

void FrHeader::
Dump( ) const
{
  Log << "Bytes: 0-4:   " << m_data.base.igwd << std::endl
      << "Bytes: 5:     " << int(m_data.base.version) << std::endl
    ;
  switch( m_type )
  {
  case T1:
    Log
      << "Bytes: 6:     " << int(m_data.t1.minorVersion) << std::endl
      << "Bytes: 7:     " << int(m_data.t1.sizeInt2) << std::endl
      << "Bytes: 8:     " << int(m_data.t1.sizeInt4) << std::endl
      << "Bytes: 9:     " << int(m_data.t1.sizeInt8) << std::endl
      << "Bytes: 10:    " << int(m_data.t1.sizeReal4) << std::endl
      << "Bytes: 11:    " << int(m_data.t1.sizeReal8) << std::endl
      << "Bytes: 38-39: " << m_data.t1.AZ[0] << m_data.t1.AZ[1] << std::endl
      ;
    
    break;
  }
}

std::string FrHeader::
String( const std::string& Class, void* Data ) const
{
  std::ostringstream	retval;

#define CONVERT( T ) \
  else if ( Class.compare( #T ) == 0 )		\
  { \
    T* d = static_cast< T* >( Data ); \
    retval << (*d);		      \
  }

  if ( Data == (void*)NULL )
  {
    retval << "NULL";
  }
  else if ( string_type::Match( Class ) )
  {
    retval << static_cast< string_type*>( Data )->Print( );
  }
  CONVERT( CHAR )
  CONVERT( CHAR_U )
  CONVERT( INT_2U )
  CONVERT( INT_2S )
  CONVERT( INT_4U )
  CONVERT( INT_4S )
  CONVERT( INT_8U )
  CONVERT( INT_8S )
  CONVERT( REAL_4 )
  CONVERT( REAL_8 )
  else if ( ptr_struct_type::Match( Class ) )
  {
    retval << static_cast< ptr_struct_type* >( Data )->Print( );
  }
  else if ( array_type::Match( Class ) )
  {
    retval << static_cast< array_type* >( Data )->Print( );
  }
  else
  {
    retval << "Unknown class: " << Class;
  }
#undef CONVERT
  return retval.str( );
}

const Description& FrHeader::
GetDescription( class_id_type Id ) const
{
  structs_type::const_iterator	cur = m_structs.find( Id );

  if ( cur == m_structs.end( ) )
  {
    throw std::range_error( "Object description does not exist" );
  }
  return cur->second;
}


void* FrHeader::
Read( const std::string& Class,
      const SymbolTable& SymbolTable )
{
  void* retval = (void*)NULL;
  INT_2U	ver = Version( );

#define READER( T ) \
  else if ( Class == #T ) \
  { \
    retval = new T; \
    FStream.Read( retval, sizeof( T ) ); \
    if ( m_byte_swapping ) \
    { \
      reverse< sizeof( T ) >( retval, 1 ); \
    } \
  }

  if ( string_type::Match( Class ) )
  {
    retval = string_type::Read( ver );
  }
  READER( CHAR )
  READER( CHAR_U )
  READER( INT_2U )
  READER( INT_2S )
  READER( INT_4U )
  READER( INT_4S )
  READER( INT_8U )
  READER( INT_8S )
  READER( REAL_4 )
  READER( REAL_8 )
  else if ( ptr_struct_type::Match( Class ) )
  {
    return ptr_struct_type::Read( ver );
  }
  else if ( array_type::Match( Class ) )
  {
    return array_type::Read( ver, Class, SymbolTable );
  }
  else
  {
    throw std::runtime_error( "Unable to read class" );
  }

#undef READER

  return retval;
}

//=======================================================================
// FrObject class definitions
//=======================================================================

FrObject::
FrObject( )
  : m_start_pos( FStream.Position( ) )
{
  if ( Header.Version( ) >= 5 )
  {
    m_core.reset( new FrBase< INT_8U, INT_2U, INT_4U >( Header ) );
  }
  else if ( Header.Version( ) >= 3 )
  {
    m_core.reset( new FrBase< INT_4U, INT_2U, INT_2U >( Header ) );
  }
  else
  {
    m_core.reset( (FrBaseCore*)NULL );
  }
  Read( );
}

FrObject::
~FrObject( )
{
}

void FrObject::
AppendSE( Description& D ) const
{
  D.AddSE( Header.String( "STRING", m_data[0] ),
	   Header.String( "STRING", m_data[1] ),
	   Header.String( "STRING", m_data[2] ) );
}

class_id_type FrObject::
StructId( ) const
{
  return m_core->Class( );
}

class_id_type FrObject::
ClassId( ) const
{
  class_id_type retval = 0;

  if ( m_core->Class( ) == 1 )
  {
    const Description& d = Header.GetDescription( 1 );

    std::string	str( Header.String( d.Class( 1 ), m_data[ 1 ]) );

    std::istringstream	s( Header.String( d.Class( 1 ), m_data[ 1 ]) );

    s >> retval;
  }

  return retval;
}

std::string FrObject::
ClassName( ) const
{
  std::string retval;

  if ( StructId( ) == 1 )
  {
    return Header.String( "STRING", m_data[0] );
  }

  return retval;
}

void FrObject::
Dump( ) const
{
  Log << "Position: " << m_start_pos << " ";
  if ( m_core.get( ) )
  {
    m_core->Dump( );
    try
    {
      const Description&
	d = Header.GetDescription( m_core->Class( ) );

      d.Dump( m_data );
    }
    catch( ... )
    {
    }
  }
  else
  {
    Log << "" << std::endl;
  }
}

void FrObject::
Read( )
{
  if ( m_core.get( ) )
  {
    Header.AppendDescription( *this, true );
    try
    {
      // Need to read the object
      const Description&	d = Header.GetDescription( m_core->Class( ) );
      INT_8U end = FStream.Position( )
	+ ( m_core->Length( ) - m_core->Bytes( ) );

      bool complete = d.Read( Header, m_data );
      if ( ! complete )
      {
	// If the object has not been defined, then advance to the start
	//   of the next object
	FStream.AdvanceABS( end );
      }
    }
    catch( ... )
    {
      // If the object has not been defined, then advance to the start
      //   of the next object
      FStream.Advance( m_core->Length( ) - m_core->Bytes( ) );
    }
    if ( ( m_core->Class( ) == 1 )
	 || ( m_core->Class( ) == 2 ) )
    {
      Header.AppendDescription( *this, false );
    }
  }
}

//=======================================================================
//=======================================================================
FrObject::FrBaseCore::
~FrBaseCore( )
{
}

//=======================================================================
//=======================================================================
template< typename LT, typename CT, typename IT >
FrObject::FrBase< LT, CT, IT >::
FrBase( FrHeader& Header )
{
  FrameStream::position_type	here( FStream.Position( ) );

  FStream.Read( &m_length, sizeof( m_length ) );
  FStream.Read( &m_class, sizeof( m_class ) );
  FStream.Read( &m_instance, sizeof( m_instance ) );
  if ( here != FStream.Position( ) )
  {
    if ( Header.ByteSwapping( ) )
    {
      reverse< sizeof( m_length ) >( &m_length , 1 );
      reverse< sizeof( m_class ) >( &m_class , 1 );
      reverse< sizeof( m_instance ) >( &m_instance , 1 );
    }
  }
}

template< typename LT, typename CT, typename IT >
void FrObject::FrBase< LT, CT, IT >::
Dump( ) const
{
  std::string	c( "Unknown" );
  try
  {
    const Description& d = Header.GetDescription( m_class );

    c = d.ClassName( );
  }
  catch( ... )
  {
  }

  Log << "Class: " << m_class << " (" << c << ")"
      << " Instance: " << m_instance
      << " Length: " << m_length
      << std::endl;
}

template< typename LT, typename CT, typename IT >
FrObject::FrBaseCore::size_type FrObject::FrBase< LT, CT, IT >::
Bytes( ) const
{
  return
    sizeof( m_length )
    + sizeof( m_class )
    + sizeof( m_instance )
    ;
}

template< typename LT, typename CT, typename IT >
class_id_type FrObject::FrBase< LT, CT, IT >::
Class( ) const
{
  return m_class;
}

template< typename LT, typename CT, typename IT >
FrObject::FrBaseCore::size_type FrObject::FrBase< LT, CT, IT >::
Length( ) const
{
  return m_length;
}

//=======================================================================
// Output class definition
//=======================================================================

Output::
Output( std::ostream& Stream )
  : m_stream( Stream )
{
}

template< class T >
std::ostream& Output::
operator<<( const T& Data )
{
  m_stream << Data;
  return m_stream;
}

//=======================================================================
// array_type
//=======================================================================
bool FrHeader::array_type::
Match( const std::string& Class )
{
  RegexMatch	m(3);

  if ( m.match( pattern, Class.c_str( ) ) )
  {
    return true;
  }
  return false;
}

std::string FrHeader::array_type::
Print( ) const
{
  return "...";
}

FrHeader::array_type* FrHeader::array_type::
Read( const version_type Version, const std::string& Class,
      const SymbolTable& Symbols )
{
  RegexMatch m( 3 );
 
  if ( m.match( pattern, Class.c_str( ) ) )
  {
    const std::string t( m.getSubString( 1 ) );
    const std::string size( m.getSubString( 2 ) );
    const INT_8U value( Symbols.LookupInteger( size ) );
    
    if ( value )
    {
      if ( string_type::Match( t ) )
      {
	for ( INT_8U count = value;
	      count > 0;
	      --count )
	{
	  std::auto_ptr< string_type > s( string_type::Read( Version ) );
	}
      } else {
	FStream.Advance( value * ClassSize( t ) );
      }
      return new array_type;
    }
    return (array_type*)NULL;
  }
  throw std::runtime_error( "Unable to read array_type" );
}

Regex FrHeader::array_type::pattern( "(CHAR|"
				     "CHAR_U|"
				     "INT_2S|"
				     "INT_2U|"
				     "INT_4S|"
				     "INT_4U|"
				     "INT_8S|"
				     "INT_8U|"
				     "REAL_4|"
				     "REAL_8|"
				     "STRING|"
				     "PTR_STRUCT|"
				     "COMPLEX_8|"
				     "COMPLEX_16)"
				     "[[:space:]]*\\[[[:space:]]*"
				     "([^[:space:]]+)"
				     "[[:space:]]*\\]", REG_EXTENDED );

FrHeader::array_type::
array_type( )
{
}

//=======================================================================
// ptr_struct_type
//=======================================================================
bool FrHeader::ptr_struct_type::
Match( const std::string& Class )
{
  return( Class.substr( 0, 10 ) == "PTR_STRUCT" );
}

std::string FrHeader::ptr_struct_type::
Print(  ) const
{
  std::ostringstream	retval;

  if ( m_version >= 3 )
  {
    retval << m_class << "-" << m_instance;
  }
  else
  {
    retval << "Unsupported version: " << m_version;
  }
  return retval.str( );
}

FrHeader::ptr_struct_type* FrHeader::ptr_struct_type::
Read( const version_type Version )
{
  if ( Version >= 6 )
  {
    INT_2U	c;
    INT_4U	i;

    FStream.Read( &c, sizeof( c ) );
    FStream.Read( &i, sizeof( i ) );

    if ( Header.ByteSwapping( ) )
    {
      reverse< sizeof( c ) >( &c, 1 );
      reverse< sizeof( i ) >( &i, 1 );
    }
    return new ptr_struct_type( c, i, Version );
  }
  else if ( Version >= 3 )
  {
    INT_2U	c;
    INT_2U	i;

    FStream.Read( &c, sizeof( c ) );
    FStream.Read( &i, sizeof( i ) );

    if ( Header.ByteSwapping( ) )
    {
      reverse< sizeof( c ) >( &c, 1 );
      reverse< sizeof( i ) >( &i, 1 );
    }
    return new ptr_struct_type( c, i, Version );
  }
  throw std::runtime_error( "Unable to read PTR_STRUCT" );
}

FrHeader::ptr_struct_type::
ptr_struct_type( INT_2U Class, INT_4U Instance, version_type Version )
  : m_class( Class ),
    m_instance( Instance ),
    m_version( Version )
{
}

//=======================================================================
// string_type
//=======================================================================
bool FrHeader::string_type::
Match( const std::string& Class )
{
  return( Class == "STRING" );
}

std::string FrHeader::string_type::
Print( ) const
{
  std::ostringstream	retval;

  if ( m_version >= 3 )
  {
    if ( m_data )
    {
      retval << m_data;
    }
  }
  else
  {
    retval << "Unsupported version: " << m_version;
  }
  return retval.str( );
}

FrHeader::string_type* FrHeader::string_type::
Read( const version_type Version )
{
  if ( Version >= 3 )
  {
    INT_2U length;
    FStream.Read( &length, sizeof( length ) );
    if ( Header.ByteSwapping( ) )
    {
      reverse< sizeof( length ) >( &length, 1 );
    }

    char*	data = (char*)NULL;

    if ( length )
    {
      data = new char[ length ];
      FStream.Read( data, length );
    }
    return new string_type( length, data, Version );
  }
  throw std::runtime_error( "Unable to read STRING" );
}

FrHeader::string_type::
string_type( INT_2U Length, const char* const Data, version_type Version )
  : m_len( Length ),
    m_data( Data ),
    m_version( Version )
{
}
