/* -*- mode: c++; c-basic-offset: 4; -*- */

#ifndef MUTEXLOCK_HH
#define MUTEXLOCK_HH

#include <assert.h>
#include <errno.h>
#include <string.h>

#include <pthread.h>

#include <iostream>

/// \class MutexLock
///
/// \brief Descibes an object which release a mutex lock
///
/// This class ensures that the given mutex is unlocked whenever an
/// exception is thrown but not caught in a locked section.
///
/// It should be used like this:
///
/// \code
/// {
///     MutexLock lock(pthread_mutex_t& mutex); // mutex is now locked
///     do whatever...
///     ...
/// } // MutexLock has been destructed so mutex is now unlocked
/// \endcode
///
///
/// Creating two MutexLock objects with the same mutex in the same scope
/// will cause a deadlock.
///
class MutexLock {
public:
    /// \brief Alias for the system specific mutex type.
    typedef pthread_mutex_t	lock_type;

    /// \brief  Constructor
    MutexLock(lock_type& mutex,
	      const char* const File = "", const unsigned int Line = 0,
	      bool Logging = true );

    /// \brief  Constructor
    MutexLock(lock_type* const mutex_ptr,
	      const char* const File = "", const unsigned int Line = 0,
	      bool Logging = true );

    /// \brief  Destructor
    ~MutexLock( );

    /// \brief  Release the lock before object is destructed
    void Release( const char* const File = "", const unsigned int Line = 0 );

    ///\brief Releases resources associated with Lock.
    inline static int Destroy( lock_type& Lock )
    {
	return pthread_mutex_destroy( &Lock );
    }

    ///\brief Allocates resources needed for lock.
    inline static lock_type Initialize( )
    {
	static const lock_type retval = PTHREAD_MUTEX_INITIALIZER;

	return retval;
    }

    /// \brief Allocate resources needed for Lock.
    inline static void Initialize( lock_type& Lock )
    {
	pthread_mutex_init( &Lock,
			    static_cast< pthread_mutexattr_t* >( NULL ) );
    }

    /// \brief Request exclusive access to Lock.
    inline static void Lock( lock_type& Lock )
    {
	pthread_mutex_lock( &Lock );
    }

    /// \brief Request exclusive access to Lock.
    inline static void Lock( lock_type* Lock )
    {
	pthread_mutex_lock( Lock );
    }

    /// \brief Relinquishes exclusive access to Lock.
    inline static void UnLock( lock_type& Lock )
    {
	pthread_mutex_unlock( &Lock );
    }

    /// \brief Relinquishes exclusive access to Lock.
    inline static void UnLock( lock_type* Lock )
    {
	pthread_mutex_unlock( Lock );
    }

private:
    /// \brief  Default Constructor
    ///
    /// Default constructor is private to prevent copying of classes
    /// internal data.
    MutexLock();

    /// \brief  Copy Constructor
    ///
    /// Copy constructor is private to prevent copying of classes
    ///   internal data.
    MutexLock(const MutexLock&);

    /// \brief  Assigment operator
    ///
    /// Assignment operator is privat to prevent copying of classes internal
    /// data
    const MutexLock& operator=(const MutexLock&);

    /// \brief  Object to prevent multiple access to critical sections of code.
    lock_type* m_mutex_ptr;

    /// \brief Keeps track of logging state.
    bool	m_logging;
};

#endif // MUTEXLOCK_HH
